@@ -1,21 +1,33 @@
 /*****************************************************************************
  * This file is part of Kvazaar HEVC encoder.
  *
- * Copyright (C) 2013-2015 Tampere University of Technology and others (see
- * COPYING file).
- *
- * Kvazaar is free software: you can redistribute it and/or modify it under
- * the terms of the GNU Lesser General Public License as published by the
- * Free Software Foundation; either version 2.1 of the License, or (at your
- * option) any later version.
- *
- * Kvazaar is distributed in the hope that it will be useful, but WITHOUT ANY
- * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
- * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
- * more details.
- *
- * You should have received a copy of the GNU General Public License along
- * with Kvazaar.  If not, see <http://www.gnu.org/licenses/>.
+ * Copyright (c) 2021, Tampere University, ITU/ISO/IEC, project contributors
+ * All rights reserved.
+ * 
+ * Redistribution and use in source and binary forms, with or without modification,
+ * are permitted provided that the following conditions are met:
+ * 
+ * * Redistributions of source code must retain the above copyright notice, this
+ *   list of conditions and the following disclaimer.
+ * 
+ * * Redistributions in binary form must reproduce the above copyright notice, this
+ *   list of conditions and the following disclaimer in the documentation and/or
+ *   other materials provided with the distribution.
+ * 
+ * * Neither the name of the Tampere University or ITU/ISO/IEC nor the names of its
+ *   contributors may be used to endorse or promote products derived from
+ *   this software without specific prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+ * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+ * INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION HOWEVER CAUSED AND ON
+ * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * INCLUDING NEGLIGENCE OR OTHERWISE ARISING IN ANY WAY OUT OF THE USE OF THIS
  ****************************************************************************/
 
 #include "cfg.h"
@@ -73,6 +85,7 @@ int kvz_config_init(kvz_config *cfg)
   cfg->vui.chroma_loc  = 0; /* left center */
   cfg->aud_enable      = 0;
   cfg->cqmfile         = NULL;
+  cfg->fast_coeff_table_fn = NULL;
   cfg->ref_frames      = 1;
   cfg->gop_len         = 4;
   cfg->gop_lowdelay    = true;
@@ -126,9 +139,9 @@ int kvz_config_init(kvz_config *cfg)
   cfg->gop_lp_definition.t = 1;
   cfg->open_gop = true;
 
-  cfg->roi.width = 0;
-  cfg->roi.height = 0;
-  cfg->roi.dqps = NULL;
+  cfg->roi.file_path = NULL;
+  cfg->roi.format = KVZ_ROI_TXT;
+
   cfg->set_qp_in_cu = false;
 
   cfg->erp_aqp = false;
@@ -163,6 +176,18 @@ int kvz_config_init(kvz_config *cfg)
   cfg->intra_bit_allocation = false;
   cfg->clip_neighbour = true;
 
+  cfg->file_format = KVZ_FORMAT_AUTO;
+
+  cfg->stats_file_prefix = NULL;
+
+  cfg->fastrd_sampling_on = 0;
+  cfg->fastrd_accuracy_check_on = 0;
+  cfg->fastrd_learning_outdir_fn = NULL;
+
+  cfg->combine_intra_cus = 1;
+  cfg->force_inter = 0;
+  cfg->intra_chroma_search = 0;
+  cfg->fast_bipred = 1;
   return 1;
 }
 
@@ -170,11 +195,13 @@ int kvz_config_destroy(kvz_config *cfg)
 {
   if (cfg) {
     FREE_POINTER(cfg->cqmfile);
+    FREE_POINTER(cfg->roi.file_path);
+    FREE_POINTER(cfg->fast_coeff_table_fn);
     FREE_POINTER(cfg->tiles_width_split);
     FREE_POINTER(cfg->tiles_height_split);
     FREE_POINTER(cfg->slice_addresses_in_ts);
-    FREE_POINTER(cfg->roi.dqps);
     FREE_POINTER(cfg->optional_key);
+    FREE_POINTER(cfg->fastrd_learning_outdir_fn);
   }
   free(cfg);
 
@@ -448,8 +475,10 @@ int kvz_config_parse(kvz_config *cfg, const char *name, const char *value)
   static const char * const scaling_list_names[] = { "off", "custom", "default", NULL };
 
   static const char * const rc_algorithm_names[] = { "no-rc", "lambda", "oba", NULL };
-  static const char * const preset_values[11][25*2] = {
 
+  static const char * const file_format_names[] = {"auto", "y4m", "yuv", NULL};
+
+  static const char * const preset_values[11][25*2] = {
       {
         "ultrafast",
         "rd", "0",
@@ -676,7 +705,7 @@ int kvz_config_parse(kvz_config *cfg, const char *name, const char *value)
       },
       {
         "veryslow",
-        "rd", "2",
+        "rd", "3",
         "pu-depth-intra", "1-4",
         "pu-depth-inter", "0-3",
         "me", "tz",
@@ -704,7 +733,7 @@ int kvz_config_parse(kvz_config *cfg, const char *name, const char *value)
       },
       {
         "placebo",
-        "rd", "2",
+        "rd", "3",
         "pu-depth-intra", "1-4",
         "pu-depth-inter", "0-3",
         "me", "tz",
@@ -850,6 +879,30 @@ int kvz_config_parse(kvz_config *cfg, const char *name, const char *value)
     cfg->cqmfile = cqmfile;
     cfg->scaling_list = KVZ_SCALING_LIST_CUSTOM;
   }
+  else if OPT("fast-coeff-table") {
+    char* fast_coeff_table_fn = strdup(value);
+    if (!fast_coeff_table_fn) {
+      fprintf(stderr, "Failed to allocate memory for fast coeff table file name.\n");
+      return 0;
+    }
+    FREE_POINTER(cfg->fast_coeff_table_fn);
+    cfg->fast_coeff_table_fn = fast_coeff_table_fn;
+  }
+  else if OPT("fastrd-sampling") {
+    cfg->fastrd_sampling_on = 1;
+  }
+  else if OPT("fastrd-accuracy-check") {
+    cfg->fastrd_accuracy_check_on = 1;
+  }
+  else if OPT("fastrd-outdir") {
+    char *fastrd_learning_outdir_fn = strdup(value);
+    if (!fastrd_learning_outdir_fn) {
+      fprintf(stderr, "Failed to allocate memory for fast RD learning outfile name.\n");
+      return 0;
+    }
+    FREE_POINTER(cfg->fastrd_learning_outdir_fn);
+    cfg->fastrd_learning_outdir_fn = fastrd_learning_outdir_fn;
+  }
   else if OPT("scaling-list") {    
     int8_t scaling_list = KVZ_SCALING_LIST_OFF;
     int result = parse_enum(value, scaling_list_names, &scaling_list);
@@ -1193,60 +1246,29 @@ int kvz_config_parse(kvz_config *cfg, const char *name, const char *value)
   }
   else if OPT("implicit-rdpcm")
     cfg->implicit_rdpcm = (bool)atobool(value);
+
   else if OPT("roi") {
-    // The ROI description is as follows:
-    // First number is width, second number is height,
-    // then follows width * height number of dqp values.
-    FILE* f = fopen(value, "rb");
-    if (!f) {
-      fprintf(stderr, "Could not open ROI file.\n");
+    static enum kvz_roi_format const formats[] = { KVZ_ROI_TXT, KVZ_ROI_BIN };
+    static const char * const format_names[] = { "txt", "bin", NULL };
+
+    char *roi_file = strdup(value);
+    if (!roi_file) {
+      fprintf(stderr, "Failed to allocate memory for ROI file name.\n");
       return 0;
     }
+    FREE_POINTER(cfg->roi.file_path);
+    cfg->roi.file_path = roi_file;
 
-    int width = 0;
-    int height = 0;
-    if (!fscanf(f, "%d", &width) || !fscanf(f, "%d", &height)) {
-      fprintf(stderr, "Failed to read ROI size.\n");
-      fclose(f);
-      return 0;
+    // Get file extension or the substring after the last dot
+    char *maybe_extension = strrchr(cfg->roi.file_path, '.');
+    if (!maybe_extension) {
+      cfg->roi.format = KVZ_ROI_TXT;
+    } else {
+      maybe_extension++;
+      int8_t format;
+      bool unknown_format = !parse_enum(maybe_extension, format_names, &format);
+      cfg->roi.format = unknown_format ? KVZ_ROI_TXT : formats[format];
     }
-
-    if (width <= 0 || height <= 0) {
-      fprintf(stderr, "Invalid ROI size: %dx%d.\n", width, height);
-      fclose(f);
-      return 0;
-    }
-
-    if (width > 10000 || height > 10000) {
-      fprintf(stderr, "ROI dimensions exceed arbitrary value of 10000.\n");
-      fclose(f);
-      return 0;
-    }
-
-    const unsigned size = width * height;
-    int8_t *dqp_array  = calloc((size_t)size, sizeof(cfg->roi.dqps[0]));
-    if (!dqp_array) {
-      fprintf(stderr, "Failed to allocate memory for ROI table.\n");
-      fclose(f);
-      return 0;
-    }
-
-    FREE_POINTER(cfg->roi.dqps);
-    cfg->roi.dqps   = dqp_array;
-    cfg->roi.width  = width;
-    cfg->roi.height = height;
-
-    for (int i = 0; i < size; ++i) {
-      int number; // Need a pointer to int for fscanf
-      if (fscanf(f, "%d", &number) != 1) {
-        fprintf(stderr, "Reading ROI file failed.\n");
-        fclose(f);
-        return 0;
-      }
-      dqp_array[i] = CLIP(-51, 51, number);
-    }
-
-    fclose(f);
   }
   else if OPT("set-qp-in-cu") {
     cfg->set_qp_in_cu = (bool)atobool(value);
@@ -1359,6 +1381,32 @@ int kvz_config_parse(kvz_config *cfg, const char *name, const char *value)
   else if OPT("clip-neighbour") {
     cfg->clip_neighbour = atobool(value);
   }
+  else if OPT("input-file-format") {
+    int8_t file_format = 0;
+    if (!parse_enum(value, file_format_names, &file_format)) {
+      fprintf(stderr, "Invalid input file format %s. Valid values include %s, %s, and %s\n", value,
+        file_format_names[0],
+        file_format_names[1], 
+        file_format_names[2]);
+      return 0;
+    }
+    cfg->file_format = file_format;
+  }
+  else if OPT("stats-file-prefix") {
+    cfg->stats_file_prefix = strdup(value);
+  }
+  else if OPT("combine-intra-cus") {
+    cfg->combine_intra_cus = atobool(value);
+  }
+  else if OPT("force-inter") {
+    cfg->force_inter = atobool(value);
+  }
+  else if OPT("intra-chroma-search") {
+    cfg->intra_chroma_search = atobool(value);
+  }
+  else if OPT("fast-bipred") {
+    cfg->fast_bipred = atobool(value);
+  }
   else {
     return 0;
   }
@@ -1550,8 +1598,8 @@ int kvz_config_validate(const kvz_config *const cfg)
     error = 1;
   }
 
-  if (cfg->rdo < 0 || cfg->rdo > 3) {
-    fprintf(stderr, "Input error: --rd parameter out of range [0..3]\n");
+  if (cfg->rdo < 0 || cfg->rdo > 6) {
+    fprintf(stderr, "Input error: --rd parameter out of range [0..5]\n");
     error = 1;
   }
 
