@@ -1,21 +1,33 @@
 /*****************************************************************************
  * This file is part of Kvazaar HEVC encoder.
  *
- * Copyright (C) 2013-2015 Tampere University of Technology and others (see
- * COPYING file).
- *
- * Kvazaar is free software: you can redistribute it and/or modify it under
- * the terms of the GNU Lesser General Public License as published by the
- * Free Software Foundation; either version 2.1 of the License, or (at your
- * option) any later version.
- *
- * Kvazaar is distributed in the hope that it will be useful, but WITHOUT ANY
- * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
- * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
- * more details.
- *
- * You should have received a copy of the GNU General Public License along
- * with Kvazaar.  If not, see <http://www.gnu.org/licenses/>.
+ * Copyright (c) 2021, Tampere University, ITU/ISO/IEC, project contributors
+ * All rights reserved.
+ * 
+ * Redistribution and use in source and binary forms, with or without modification,
+ * are permitted provided that the following conditions are met:
+ * 
+ * * Redistributions of source code must retain the above copyright notice, this
+ *   list of conditions and the following disclaimer.
+ * 
+ * * Redistributions in binary form must reproduce the above copyright notice, this
+ *   list of conditions and the following disclaimer in the documentation and/or
+ *   other materials provided with the distribution.
+ * 
+ * * Neither the name of the Tampere University or ITU/ISO/IEC nor the names of its
+ *   contributors may be used to endorse or promote products derived from
+ *   this software without specific prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+ * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+ * INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION HOWEVER CAUSED AND ON
+ * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * INCLUDING NEGLIGENCE OR OTHERWISE ARISING IN ANY WAY OUT OF THE USE OF THIS
  ****************************************************************************/
 
 #include "rate_control.h"
@@ -27,13 +39,19 @@
 #include "pthread.h"
 
 
-static const int SMOOTHING_WINDOW = 40;
+static const int MIN_SMOOTHING_WINDOW = 40;
+static int smoothing_window = 40;
 static const double MIN_LAMBDA    = 0.1;
 static const double MAX_LAMBDA    = 10000;
 #define BETA1 1.2517
 
 static kvz_rc_data *data;
 
+static FILE *dist_file;
+static FILE *bits_file;
+static FILE *qp_file;
+static FILE *lambda_file;
+
 /**
  * \brief Clip lambda value to a valid range.
  */
@@ -86,6 +104,17 @@ kvz_rc_data * kvz_get_rc_data(const encoder_control_t * const encoder) {
 
   data->intra_alpha = 6.7542000000000000;
   data->intra_beta = 1.7860000000000000;
+  if(encoder->cfg.stats_file_prefix) {
+    char buff[128];
+    sprintf(buff, "%sbits.txt", encoder->cfg.stats_file_prefix);
+    bits_file = fopen(buff, "w");
+    sprintf(buff, "%sdist.txt", encoder->cfg.stats_file_prefix);
+    dist_file = fopen(buff, "w");
+    sprintf(buff, "%sqp.txt", encoder->cfg.stats_file_prefix);
+    qp_file = fopen(buff, "w");
+    sprintf(buff, "%slambda.txt", encoder->cfg.stats_file_prefix);
+    lambda_file = fopen(buff, "w");
+  }
   return data;
 }
 
@@ -147,21 +176,31 @@ static double gop_allocate_bits(encoder_state_t * const state)
   // At this point, total_bits_coded of the current state contains the
   // number of bits written encoder->owf frames before the current frame.
   uint64_t bits_coded = state->frame->total_bits_coded;
-  int pictures_coded = MAX(0, state->frame->num - encoder->cfg.owf);
-
-  int gop_offset = (state->frame->gop_offset - encoder->cfg.owf) % MAX(1, encoder->cfg.gop_len);
-  
-  if (encoder->cfg.gop_len > 0 && gop_offset != encoder->cfg.gop_len - 1 && encoder->cfg.gop_lp_definition.d == 0) {
+  int pictures_coded = 0;
+  if(encoder->cfg.gop_len) {
+    pictures_coded = MAX(0, state->frame->num - CEILDIV(encoder->cfg.owf, encoder->cfg.gop_len)*encoder->cfg.gop_len);
+  }
+  else {
+    pictures_coded = MAX(0, state->frame->num - encoder->cfg.owf);
+  }
+    
+  if (encoder->cfg.gop_len > 0 && encoder->cfg.owf > 0) {
     // Subtract number of bits in the partially coded GOP.
     bits_coded -= state->frame->cur_gop_bits_coded;
-    // Subtract number of pictures in the partially coded GOP.
-    pictures_coded -= gop_offset + 1;
   }
 
-  // Equation 12 from https://doi.org/10.1109/TIP.2014.2336550
-  double gop_target_bits =
-    (encoder->target_avg_bppic * (pictures_coded + SMOOTHING_WINDOW) - bits_coded)
-    * MAX(1, encoder->cfg.gop_len) / SMOOTHING_WINDOW;
+  smoothing_window = MAX(MIN_SMOOTHING_WINDOW, smoothing_window - encoder->cfg.gop_len / 2);
+  double gop_target_bits = -1;
+
+  while( gop_target_bits < 0 && smoothing_window < 150) {
+    // Equation 12 from https://doi.org/10.1109/TIP.2014.2336550
+    gop_target_bits =
+      (encoder->target_avg_bppic * (pictures_coded + smoothing_window) - bits_coded)
+      * MAX(1, encoder->cfg.gop_len) / smoothing_window;
+    if(gop_target_bits < 0) {
+      smoothing_window += 10;
+    }
+  }
   // Allocate at least 200 bits for each GOP like HM does.
   return MAX(200, gop_target_bits);
 }
@@ -773,13 +812,16 @@ static double qp_to_lambda(encoder_state_t* const state, int qp)
     int id = lcu.x + lcu.y * state->tile->frame->width_in_lcu;
     int aq_offset = round(state->frame->aq_offsets[id]);
     state->qp += aq_offset;
-    // Maximum delta QP is clipped between [-26, 25] according to ITU T-REC-H.265 specification chapter 7.4.9.10 Transform unit semantics
+    // Maximum delta QP is clipped according to ITU T-REC-H.265 specification chapter 7.4.9.10 Transform unit semantics
+    // Clipping range is a function of bit depth
     // Since this value will be later combined with qp_pred, clip to half of that instead to be safe
-    state->qp = CLIP(state->frame->QP - 13, state->frame->QP + 12, state->qp);
+    state->qp = CLIP(state->frame->QP + KVZ_QP_DELTA_MIN / 2, state->frame->QP + KVZ_QP_DELTA_MAX / 2, state->qp);
     state->qp = CLIP_TO_QP(state->qp);
     state->lambda = qp_to_lambda(state, state->qp);
     state->lambda_sqrt = sqrt(state->lambda);
 
+    ctu->adjust_lambda = state->lambda;
+    ctu->adjust_qp = state->qp;
     //ctu->qp = state->qp;
     //ctu->lambda = state->lambda;
   }
@@ -843,6 +885,22 @@ static void update_ck(encoder_state_t * const state, int ctu_index, int layer)
 }
 
 
+static int calc_poc(encoder_state_t * const state) {
+  const encoder_control_t * const encoder = state->encoder_control;
+  if((encoder->cfg.open_gop && !encoder->cfg.gop_lowdelay) || !encoder->cfg.intra_period) {
+    return state->frame->poc;
+  }
+  if(!encoder->cfg.gop_len || encoder->cfg.open_gop || encoder->cfg.intra_period == 1 || encoder->cfg.gop_lowdelay) {
+    return state->frame->poc + state->frame->num / encoder->cfg.intra_period * encoder->cfg.intra_period;
+  }
+  if (!encoder->cfg.gop_lowdelay && !encoder->cfg.open_gop) {
+    return state->frame->poc + state->frame->num / (encoder->cfg.intra_period + 1) * (encoder->cfg.intra_period + 1);
+  }
+  assert(0);
+  return -1;
+}
+
+
 void kvz_update_after_picture(encoder_state_t * const state) {
   double total_distortion = 0;
   double lambda = 0;
@@ -864,6 +922,14 @@ void kvz_update_after_picture(encoder_state_t * const state) {
     pthread_mutex_unlock(&state->frame->new_ratecontrol->intra_lock);
   }
 
+  if (encoder->cfg.stats_file_prefix) {
+    int poc = calc_poc(state);
+    fprintf(dist_file, "%d %d %d\n", poc, encoder->in.width_in_lcu, encoder->in.height_in_lcu);
+    fprintf(bits_file, "%d %d %d\n", poc, encoder->in.width_in_lcu, encoder->in.height_in_lcu);
+    fprintf(qp_file, "%d %d %d\n", poc, encoder->in.width_in_lcu, encoder->in.height_in_lcu);
+    fprintf(lambda_file, "%d %d %d\n", poc, encoder->in.width_in_lcu, encoder->in.height_in_lcu);
+  }
+
   for(int y_ctu = 0; y_ctu < state->encoder_control->in.height_in_lcu; y_ctu++) {
     for (int x_ctu = 0; x_ctu < state->encoder_control->in.width_in_lcu; x_ctu++) {
       int ctu_distortion = 0;
@@ -878,9 +944,23 @@ void kvz_update_after_picture(encoder_state_t * const state) {
       ctu->distortion = (double)ctu_distortion / ctu->pixels;
       total_distortion += (double)ctu_distortion / ctu->pixels;
       lambda += ctu->lambda / (state->encoder_control->in.width_in_lcu * state->encoder_control->in.height_in_lcu);
-    }    
+      if(encoder->cfg.stats_file_prefix) {
+        fprintf(dist_file, "%f ", ctu->distortion);
+        fprintf(bits_file, "%d ", ctu->bits);
+        fprintf(qp_file, "%d ", ctu->adjust_qp ? ctu->adjust_qp : ctu->qp);
+        fprintf(lambda_file, "%f ", ctu->adjust_lambda ? ctu->adjust_lambda : ctu->lambda);
+      }
+    }
+    if (encoder->cfg.stats_file_prefix) {
+      fprintf(dist_file, "\n");
+      fprintf(bits_file, "\n");
+      fprintf(qp_file, "\n");
+      fprintf(lambda_file, "\n");
+    }
   }
 
+  if(encoder->cfg.stats_file_prefix && encoder->cfg.rc_algorithm != KVZ_OBA) return;
+
   total_distortion /= (state->encoder_control->in.height_in_lcu * state->encoder_control->in.width_in_lcu);
   if (state->frame->is_irap) {
     pthread_mutex_lock(&state->frame->new_ratecontrol->intra_lock);
@@ -1003,25 +1083,27 @@ void kvz_set_lcu_lambda_and_qp(encoder_state_t * const state,
                                vector2d_t pos)
 {
   const encoder_control_t * const ctrl = state->encoder_control;
+  lcu_stats_t *lcu = kvz_get_lcu_stats(state, pos.x, pos.y);
 
-  if (ctrl->cfg.roi.dqps != NULL) {
-    vector2d_t lcu = {
+  if (state->tile->frame->source->roi.roi_array) {
+    vector2d_t lcu_vec = {
       pos.x + state->tile->lcu_offset_x,
       pos.y + state->tile->lcu_offset_y
     };
     vector2d_t roi = {
-      lcu.x * ctrl->cfg.roi.width / ctrl->in.width_in_lcu,
-      lcu.y * ctrl->cfg.roi.height / ctrl->in.height_in_lcu
+      lcu_vec.x * state->tile->frame->source->roi.width / ctrl->in.width_in_lcu,
+      lcu_vec.y * state->tile->frame->source->roi.height / ctrl->in.height_in_lcu
     };
-    int roi_index = roi.x + roi.y * ctrl->cfg.roi.width;
-    int dqp = ctrl->cfg.roi.dqps[roi_index];
+    int roi_index = roi.x + roi.y * state->tile->frame->source->roi.width;
+    int dqp = state->tile->frame->source->roi.roi_array[roi_index];
+    if(dqp != 0) {
+      pos.x = 0;
+    }
     state->qp = CLIP_TO_QP(state->frame->QP + dqp);
     state->lambda = qp_to_lambda(state, state->qp);
     state->lambda_sqrt = sqrt(state->lambda);
-
   }
   else if (ctrl->cfg.target_bitrate > 0) {
-    lcu_stats_t *lcu         = kvz_get_lcu_stats(state, pos.x, pos.y);
     const uint32_t pixels    = MIN(LCU_WIDTH, state->tile->frame->width  - LCU_WIDTH * pos.x) *
                                MIN(LCU_WIDTH, state->tile->frame->height - LCU_WIDTH * pos.y);
 
@@ -1054,7 +1136,6 @@ void kvz_set_lcu_lambda_and_qp(encoder_state_t * const state,
                   lambda);
     lambda = clip_lambda(lambda);
 
-    lcu->lambda        = lambda;
     state->lambda      = lambda;
     state->lambda_sqrt = sqrt(lambda);
     state->qp          = lambda_to_qp(lambda);
@@ -1065,20 +1146,27 @@ void kvz_set_lcu_lambda_and_qp(encoder_state_t * const state,
     state->lambda_sqrt = sqrt(state->frame->lambda);
   }
 
+  lcu->lambda = state->lambda;
+  lcu->qp = state->qp;
+
   // Apply variance adaptive quantization
   if (ctrl->cfg.vaq) {
-    vector2d_t lcu = {
+    vector2d_t lcu_pos = {
       pos.x + state->tile->lcu_offset_x,
       pos.y + state->tile->lcu_offset_y
     };
-    int id = lcu.x + lcu.y * state->tile->frame->width_in_lcu;
+    int id = lcu_pos.x + lcu_pos.y * state->tile->frame->width_in_lcu;
     int aq_offset = round(state->frame->aq_offsets[id]);
     state->qp += aq_offset;    
-    // Maximum delta QP is clipped between [-26, 25] according to ITU T-REC-H.265 specification chapter 7.4.9.10 Transform unit semantics
+    // Maximum delta QP is clipped according to ITU T-REC-H.265 specification chapter 7.4.9.10 Transform unit semantics
+    // Clipping range is a function of bit depth
     // Since this value will be later combined with qp_pred, clip to half of that instead to be safe
-    state->qp = CLIP(state->frame->QP - 13, state->frame->QP + 12, state->qp);
+    state->qp = CLIP(state->frame->QP + KVZ_QP_DELTA_MIN / 2, state->frame->QP + KVZ_QP_DELTA_MAX / 2, state->qp);
     state->qp = CLIP_TO_QP(state->qp);
     state->lambda = qp_to_lambda(state, state->qp);
     state->lambda_sqrt = sqrt(state->lambda);
+
+    lcu->adjust_lambda = state->lambda;
+    lcu->adjust_qp = state->qp;
   }
 }
