@@ -1,22 +1,34 @@
 /*****************************************************************************
-* This file is part of Kvazaar HEVC encoder.
-*
-* Copyright (C) 2013-2015 Tampere University of Technology and others (see
-* COPYING file).
-*
-* Kvazaar is free software: you can redistribute it and/or modify it under
-* the terms of the GNU Lesser General Public License as published by the
-* Free Software Foundation; either version 2.1 of the License, or (at your
-* option) any later version.
-*
-* Kvazaar is distributed in the hope that it will be useful, but WITHOUT ANY
-* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-* FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
-* more details.
-*
-* You should have received a copy of the GNU General Public License along
-* with Kvazaar.  If not, see <http://www.gnu.org/licenses/>.
-****************************************************************************/
+ * This file is part of Kvazaar HEVC encoder.
+ *
+ * Copyright (c) 2021, Tampere University, ITU/ISO/IEC, project contributors
+ * All rights reserved.
+ * 
+ * Redistribution and use in source and binary forms, with or without modification,
+ * are permitted provided that the following conditions are met:
+ * 
+ * * Redistributions of source code must retain the above copyright notice, this
+ *   list of conditions and the following disclaimer.
+ * 
+ * * Redistributions in binary form must reproduce the above copyright notice, this
+ *   list of conditions and the following disclaimer in the documentation and/or
+ *   other materials provided with the distribution.
+ * 
+ * * Neither the name of the Tampere University or ITU/ISO/IEC nor the names of its
+ *   contributors may be used to endorse or promote products derived from
+ *   this software without specific prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+ * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+ * INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION HOWEVER CAUSED AND ON
+ * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * INCLUDING NEGLIGENCE OR OTHERWISE ARISING IN ANY WAY OUT OF THE USE OF THIS
+ ****************************************************************************/
 
 /*
 * \file
@@ -40,6 +52,7 @@
 #include "strategyselector.h"
 #include "tables.h"
 #include "transform.h"
+#include "fast_coeff_cost.h"
 
 static INLINE int32_t hsum32_8x32i(__m256i src)
 {
@@ -506,35 +519,37 @@ void kvz_quant_avx2(const encoder_state_t * const state, const coeff_t * __restr
 #undef LOG2_SCAN_SET_SIZE
 }
 
-static INLINE __m128i get_residual_4x1_avx2(const kvz_pixel *a_in, const kvz_pixel *b_in){
+#if KVZ_BIT_DEPTH == 8
+
+static INLINE __m128i get_residual_4x1_avx2(const uint8_t *a_in, const uint8_t *b_in){
   __m128i a = _mm_cvtsi32_si128(*(int32_t*)a_in);
   __m128i b = _mm_cvtsi32_si128(*(int32_t*)b_in);
   __m128i diff = _mm_sub_epi16(_mm_cvtepu8_epi16(a), _mm_cvtepu8_epi16(b) );
   return diff;
 }
 
-static INLINE __m128i get_residual_8x1_avx2(const kvz_pixel *a_in, const kvz_pixel *b_in){
+static INLINE __m128i get_residual_8x1_avx2(const uint8_t *a_in, const uint8_t *b_in){
   __m128i a = _mm_cvtsi64_si128(*(int64_t*)a_in);
   __m128i b = _mm_cvtsi64_si128(*(int64_t*)b_in);
   __m128i diff = _mm_sub_epi16(_mm_cvtepu8_epi16(a), _mm_cvtepu8_epi16(b) );
   return diff;
 }
 
-static INLINE int32_t get_quantized_recon_4x1_avx2(int16_t *residual, const kvz_pixel *pred_in){
+static INLINE int32_t get_quantized_recon_4x1_avx2(int16_t *residual, const uint8_t *pred_in){
   __m128i res = _mm_loadl_epi64((__m128i*)residual);
   __m128i pred = _mm_cvtsi32_si128(*(int32_t*)pred_in);
   __m128i rec = _mm_add_epi16(res, _mm_cvtepu8_epi16(pred));
   return _mm_cvtsi128_si32(_mm_packus_epi16(rec, rec));
 }
 
-static INLINE int64_t get_quantized_recon_8x1_avx2(int16_t *residual, const kvz_pixel *pred_in){
+static INLINE int64_t get_quantized_recon_8x1_avx2(int16_t *residual, const uint8_t *pred_in){
   __m128i res = _mm_loadu_si128((__m128i*)residual);
   __m128i pred = _mm_cvtsi64_si128(*(int64_t*)pred_in);
   __m128i rec = _mm_add_epi16(res, _mm_cvtepu8_epi16(pred));
   return _mm_cvtsi128_si64(_mm_packus_epi16(rec, rec));
 }
 
-static void get_residual_avx2(const kvz_pixel *ref_in, const kvz_pixel *pred_in, int16_t *residual, int width, int in_stride){
+static void get_residual_avx2(const uint8_t *ref_in, const uint8_t *pred_in, int16_t *residual, int width, int in_stride){
 
   __m128i diff = _mm_setzero_si128();
   switch (width) {
@@ -579,7 +594,7 @@ static void get_residual_avx2(const kvz_pixel *ref_in, const kvz_pixel *pred_in,
   }
 }
 
-static void get_quantized_recon_avx2(int16_t *residual, const kvz_pixel *pred_in, int in_stride, kvz_pixel *rec_out, int out_stride, int width){
+static void get_quantized_recon_avx2(int16_t *residual, const uint8_t *pred_in, int in_stride, uint8_t *rec_out, int out_stride, int width){
 
   switch (width) {
     case 4:
@@ -629,8 +644,8 @@ int kvz_quantize_residual_avx2(encoder_state_t *const state,
   const cu_info_t *const cur_cu, const int width, const color_t color,
   const coeff_scan_order_t scan_order, const int use_trskip,
   const int in_stride, const int out_stride,
-  const kvz_pixel *const ref_in, const kvz_pixel *const pred_in,
-  kvz_pixel *rec_out, coeff_t *coeff_out,
+  const uint8_t *const ref_in, const uint8_t *const pred_in,
+  uint8_t *rec_out, coeff_t *coeff_out,
   bool early_skip)
 {
   // Temporary arrays to pass data to and from kvz_quant and transform functions.
@@ -766,6 +781,8 @@ void kvz_dequant_avx2(const encoder_state_t * const state, coeff_t *q_coef, coef
   }
 }
 
+#endif // KVZ_BIT_DEPTH == 8
+
 static uint32_t coeff_abs_sum_avx2(const coeff_t *coeffs, const size_t length)
 {
   assert(length % 8 == 0);
@@ -788,81 +805,64 @@ static uint32_t coeff_abs_sum_avx2(const coeff_t *coeffs, const size_t length)
   return parts[0] + parts[1] + parts[2] + parts[3];
 }
 
-#define TO_Q88(f) ((int16_t)((f) * 256.0f))
-
-static uint32_t fast_coeff_cost_avx2(const coeff_t *coeff, int32_t width, int32_t qp)
+static double fast_coeff_cost_avx2(const coeff_t *coeff, int32_t width, uint64_t weights)
 {
-#define NUM_BUCKETS 5
-  static const int16_t wt_m[NUM_BUCKETS] = {
-    TO_Q88(-0.004916),
-    TO_Q88( 0.010806),
-    TO_Q88( 0.055562),
-    TO_Q88( 0.033436),
-    TO_Q88(-0.007690),
-  };
-  static const int16_t wt_c[NUM_BUCKETS] = {
-    TO_Q88( 0.172024),
-    TO_Q88( 3.421462),
-    TO_Q88( 2.879506),
-    TO_Q88( 5.585471),
-    TO_Q88( 0.256772),
-  };
+  const __m256i zero           = _mm256_setzero_si256();
+  const __m256i threes         = _mm256_set1_epi16(3);
+  const __m256i negate_hibytes = _mm256_set1_epi16(0xff00);
+  const __m128i wt_extract_los = _mm_cvtsi32_si128(0x06040200);
+  const __m128i wt_extract_his = _mm_cvtsi32_si128(0x07050301);
 
-  const __m256i zero   = _mm256_setzero_si256();
-  const __m256i threes = _mm256_set1_epi16(3);
-  const __m256i ones   = _mm256_srli_epi16(threes, 1);
-  const __m256i twos   = _mm256_slli_epi16(ones,   1);
+  __m256i lo_sum     = _mm256_setzero_si256();
+  __m256i hi_sum     = _mm256_setzero_si256();
 
-  __m256i wt[NUM_BUCKETS - 1];
-  for (int32_t i = 0; i < NUM_BUCKETS - 1; i++)
-    wt[i] = _mm256_set1_epi16(wt_m[i] * qp + wt_c[i]);
+  __m128i wts_128    = _mm_loadl_epi64 ((const __m128i *)&weights);
+  __m128i wts_lo_128 = _mm_shuffle_epi8(wts_128, wt_extract_los);
+  __m128i wts_hi_128 = _mm_shuffle_epi8(wts_128, wt_extract_his);
 
-  uint32_t wid_wt = width * (wt_m[NUM_BUCKETS - 1] * qp + wt_c[NUM_BUCKETS - 1]);
-  __m256i avx_inc = _mm256_setzero_si256();
+  __m256i wts_lo     = _mm256_broadcastsi128_si256(wts_lo_128);
+  __m256i wts_hi     = _mm256_broadcastsi128_si256(wts_hi_128);
 
-  for (int32_t i = 0; i < width * width; i += 16) {
-    __m256i curr      = _mm256_loadu_si256((__m256i *)(coeff + i));
-    __m256i curr_abs  = _mm256_abs_epi16  (curr);
-    __m256i curr_max3 = _mm256_min_epi16  (curr_abs, threes);
+  for (int i = 0; i < width * width; i += 32) {
+    __m256i curr_lo      = _mm256_loadu_si256 ((const __m256i *)(coeff + i));
+    __m256i curr_abs_lo  = _mm256_abs_epi16   (curr_lo);
+    __m256i curr_max3_lo = _mm256_min_epu16   (curr_abs_lo, threes);
 
-    __m256i curr_eq_0 = _mm256_cmpeq_epi16(curr_max3, zero);
-    __m256i curr_eq_1 = _mm256_cmpeq_epi16(curr_max3, ones);
-    __m256i curr_eq_2 = _mm256_cmpeq_epi16(curr_max3, twos);
-    __m256i curr_eq_3 = _mm256_cmpeq_epi16(curr_max3, threes);
+    // 4x4 blocks only have 16 coeffs, so handle them separately
+    __m256i curr_max3_hi;
+    if (width >= 8) {
+      __m256i curr_hi      = _mm256_loadu_si256 ((const __m256i *)(coeff + i + 16));
+      __m256i curr_abs_hi  = _mm256_abs_epi16   (curr_hi);
+              curr_max3_hi = _mm256_min_epu16   (curr_abs_hi, threes);
+              curr_max3_hi = _mm256_slli_epi16  (curr_max3_hi, 8);
+    } else {
+      // Set MSBs for high bytes if they're meaningless, so shuffles will
+      // return zeros for them
+      curr_max3_hi = negate_hibytes;
+    }
+    __m256i curr_max3    = _mm256_or_si256    (curr_max3_lo, curr_max3_hi);
+    __m256i curr_wts_lo  = _mm256_shuffle_epi8(wts_lo, curr_max3);
+    __m256i curr_wts_hi  = _mm256_shuffle_epi8(wts_hi, curr_max3);
 
-    __m256i curr_0_wt = _mm256_and_si256  (curr_eq_0, wt[0]);
-    __m256i curr_1_wt = _mm256_and_si256  (curr_eq_1, wt[1]);
-    __m256i curr_2_wt = _mm256_and_si256  (curr_eq_2, wt[2]);
-    __m256i curr_3_wt = _mm256_and_si256  (curr_eq_3, wt[3]);
+    __m256i curr_sum_lo  = _mm256_sad_epu8    (curr_wts_lo, zero);
+    __m256i curr_sum_hi  = _mm256_sad_epu8    (curr_wts_hi, zero);
 
-    // Use madd to horizontally sum 16-bit weights into 32-bit atoms
-    __m256i wt_0_32b  = _mm256_madd_epi16(curr_0_wt, ones);
-    __m256i wt_1_32b  = _mm256_madd_epi16(curr_1_wt, ones);
-    __m256i wt_2_32b  = _mm256_madd_epi16(curr_2_wt, ones);
-    __m256i wt_3_32b  = _mm256_madd_epi16(curr_3_wt, ones);
-
-    __m256i wt_01     = _mm256_add_epi32(wt_0_32b, wt_1_32b);
-    __m256i wt_23     = _mm256_add_epi32(wt_2_32b, wt_3_32b);
-    __m256i curr_wts  = _mm256_add_epi32(wt_01,    wt_23);
-    avx_inc           = _mm256_add_epi32(avx_inc,  curr_wts);
+            lo_sum       = _mm256_add_epi64   (lo_sum, curr_sum_lo);
+            hi_sum       = _mm256_add_epi64   (hi_sum, curr_sum_hi);
   }
-  __m128i inchi = _mm256_extracti128_si256(avx_inc, 1);
-  __m128i inclo = _mm256_castsi256_si128  (avx_inc);
+          hi_sum = _mm256_slli_epi64(hi_sum, 8);
+  __m256i sum0   = _mm256_add_epi64(lo_sum, hi_sum);
 
-  __m128i sum_1 = _mm_add_epi32    (inclo, inchi);
-  __m128i sum_2 = _mm_shuffle_epi32(sum_1, _MM_SHUFFLE(1, 0, 3, 2));
-  __m128i sum_3 = _mm_add_epi32    (sum_1, sum_2);
-  __m128i sum_4 = _mm_shuffle_epi32(sum_3, _MM_SHUFFLE(2, 3, 0, 1));
-  __m128i sum   = _mm_add_epi32    (sum_3, sum_4);
+  __m256i sum1   = _mm256_permute4x64_epi64(sum0, _MM_SHUFFLE(1, 0, 3, 2));
+  __m256i sum2   = _mm256_add_epi64        (sum0, sum1);
+  __m256i sum3   = _mm256_shuffle_epi32    (sum2, _MM_SHUFFLE(1, 0, 3, 2));
+  __m256i sum4   = _mm256_add_epi64        (sum2, sum3);
 
-  uint32_t sum_u32 = _mm_cvtsi128_si32(sum);
-  uint32_t sum_total = sum_u32 + wid_wt;
-  return sum_total >> 8;
-#undef NUM_BUCKETS
+  __m128i sum128 = _mm256_castsi256_si128  (sum4);
+  uint32_t temp = _mm_cvtsi128_si32(sum128);
+  return (double)(temp) / 256.0;
 }
 
-#undef TO_Q88
-
 #endif //COMPILE_INTEL_AVX2 && defined X86_64
 
 int kvz_strategy_register_quant_avx2(void* opaque, uint8_t bitdepth)
@@ -870,11 +870,13 @@ int kvz_strategy_register_quant_avx2(void* opaque, uint8_t bitdepth)
   bool success = true;
 
 #if COMPILE_INTEL_AVX2 && defined X86_64
-  success &= kvz_strategyselector_register(opaque, "quant", "avx2", 40, &kvz_quant_avx2);
+#if KVZ_BIT_DEPTH == 8
   if (bitdepth == 8) {
     success &= kvz_strategyselector_register(opaque, "quantize_residual", "avx2", 40, &kvz_quantize_residual_avx2);
     success &= kvz_strategyselector_register(opaque, "dequant", "avx2", 40, &kvz_dequant_avx2);
   }
+#endif // KVZ_BIT_DEPTH == 8
+  success &= kvz_strategyselector_register(opaque, "quant", "avx2", 40, &kvz_quant_avx2);
   success &= kvz_strategyselector_register(opaque, "coeff_abs_sum", "avx2", 0, &coeff_abs_sum_avx2);
   success &= kvz_strategyselector_register(opaque, "fast_coeff_cost", "avx2", 40, &fast_coeff_cost_avx2);
 #endif //COMPILE_INTEL_AVX2 && defined X86_64
